(* Mathematica Package *)

BeginPackage["AtomicDensityMatrix`AtomicSystem`", 
	{"AtomicDensityMatrix`UsageMessages`", 
	"AtomicDensityMatrix`Common`", 
	"AtomicDensityMatrix`AtomicDensityMatrix`"}]

Begin["`Private`"] (* Begin Private Context *) 

(* ::Subsection:: *)
(*Atomic system*)

(*AtomicState[ label ] specifies an atomic state labeled by label*)
(*AtomicState[ label, {\!\(TraditionalForm\`\**)
(*SubscriptBox[*)
(*StyleBox["param",*)
(*FontSlant->"Italic"], "1"]\)->Subscript[value, 1] , ...} ] specifies an atomic state with one or more parameters set*)
(**)
(*Level parameters:*)
(*	S		electronic spin quantum number*)
(*	L		electronic orbital angular momentum quantum number*)
(*	J		total electronic angular momentum quantum number*)
(*	NuclearSpin	nuclear spin quantum number*)
(*	F		total angular momentum quantum number*)
(*	M		z-projection of total angular momentum*)
(*	Parity		parity*)
(*	Energy		unperturbed energy*)
(*	NaturalWidth	natural line width*)
(*	BranchingRatio[label]	branching ratio to state designated by label*)
(*	LevelPosition	"temporary" parameter that specifies whether the state is upper or lower for the purposes of a spontaneous decay calculation*)

AtomicState[label_,{rules___Rule}]:=AtomicState[label,rules]

(*StateLabel[state] gets the label from a state specification*)

StateLabel[state_AtomicState] := First[state];
StateLabel[sys:{_AtomicState..}] := StateLabel/@sys;

(*StateParameters[state] gets the parameters from a state specification*)

StateParameters[state_AtomicState]^:=Rest[List@@state]

MapOnParameters[f_,state_AtomicState]:=AtomicState[StateLabel[state],f[StateParameters[state]]]

(*StateParameterPattern[] matches valid atomic state parameters*)

StateParameterPattern[] := 
 S | L | J | NuclearSpin | F | M | Parity | Energy | NaturalWidth | 
  GFactor | BranchingRatio[_] | LevelPosition | HyperfineA | 
  HyperfineB | Polarizability[_] | TransitionPolarizability[_]

(*StateParameterQ[expr] tests if an expression is a valid atomic state parameter*)

StateParameterQ[expr_]:=MatchQ[expr,StateParameterPattern[]]

(*AtomicSystemQ[expr] tests if an expression is a list of AtomicState specifications.*)

AtomicSystemQ[expr_]:=MatchQ[expr,{_AtomicState..}]

(*DefinedQ[state,param] checks if an atomic state parameter is defined in a state specification.*)

DefinedQ[state_AtomicState, X : StateParameterPattern[]] := MemberQ[Rest[state], X -> _];

DefinedQ[sys : _AtomicState | {_AtomicState ..}, X : {StateParameterPattern[] ..} | StateParameterPattern[] ..] := 
	And @@ Flatten@Outer[DefinedQ, Flatten@{sys}, Flatten@{X}];
 
(*DeleteParameter[state, param] deletes a rule from a state specification. More than one parameter can be deleted by giving a list of parameters or additional arguments.*)

DeleteParameter[sys : {_AtomicState ..} | _AtomicState, 
  X : {StateParameterPattern[] ..} | StateParameterPattern[]..] := 
 DeleteCases[sys, Alternatives @@ Thread[Flatten[{X}] -> _], {ArrayDepth[{sys}]}]
 
(*AddParameter[state, param -> value] adds a rule to an atomic state specification. If the parameter is already present, its value is changed if necessary.*)

AddParameter[state_AtomicState, rules : {(StateParameterPattern[] -> _) ..} | (StateParameterPattern[] -> _)..] := 
	Join[DeleteParameter[state, Flatten[{rules}][[All, 1]]], AtomicState @@ Flatten[{rules}]]
AddParameter[sys : {_AtomicState ..}, rules : {(StateParameterPattern[] -> _) ..} | (StateParameterPattern[] -> _)..] := 
	Thread[Unevaluated@AddParameter[sys, rules], List, 1]  

(*param[state] returns the parameter for a state if defined, otherwise returns a placeholder marked by the state label.*)
(*It caches results for improved performance.*)

(*(#[state_AtomicState]:=#/.StateParameters[state]/.#->#[StateLabel[state]])&/@{S,L,NuclearSpin,M,Parity,Energy,NaturalWidth,GFactor,LevelPosition,HyperfineA,HyperfineB};*)

With[{f=#},
	f[state_AtomicState] := f[state] = f/.StateParameters[state]/.f->f[StateLabel[state]];
	f[sys:{_AtomicState..}] := f/@sys;
]&/@{S,L,NuclearSpin,M,Parity,Energy,NaturalWidth,GFactor,LevelPosition,HyperfineA,HyperfineB};

(*J[state] returns the J for a state if defined, otherwise returns 0*)

J[state_AtomicState]:=J[state]=J/.StateParameters[state]/.J->0
J[sys:{_AtomicState..}] := J/@sys;

(*F[state] returns F for a state if defined, otherwise returns J*)

F[state_AtomicState]:=F[state]=F/.StateParameters[state]/.F->J[state]
F[sys:{_AtomicState..}] := F/@sys;

BranchingRatio[state1_AtomicState][state2_AtomicState]:=If[
	(*if reverse branching ratio is nonzero, then this b.r. is zero*)
	DefinedQ[state1,BranchingRatio[state2]]&&BranchingRatio[StateLabel[state2]][state1]=!=0, 0,
	(*else calculate b.r. using state label*)
	BranchingRatio[StateLabel[state1]][state2]
]

BranchingRatio[label_][state_AtomicState]:= Module[{value},
	If[
		StateLabel[state]===label||NaturalWidth[state]===0,
		0,
		If[
			(value=BranchingRatio[label]/.StateParameters[state])=!=BranchingRatio[label],
			value,
			BranchingRatio[StateLabel[state],label]
		]
	]
]

Polarizability[k__][state_AtomicState]:= Module[{value},
	If[(value=Polarizability[k]/.StateParameters[state])=!=Polarizability[k],
		value,
		Polarizability[k][StateLabel[state]]
	]
]

TransitionPolarizability[k_][state1_AtomicState,state2_AtomicState]:=Module[{value},
	Which[
	(value=TransitionPolarizability[k][StateLabel[state1]]/.StateParameters[state2])=!=TransitionPolarizability[k][StateLabel[state1]],value,
	(value=TransitionPolarizability[k][StateLabel[state2]]/.StateParameters[state1])=!=TransitionPolarizability[k][StateLabel[state2]],value,
	True,TransitionPolarizability[k][Sequence@@Sort[StateLabel/@{state1,state2}]]
	]
]

GroundStateQ[state_AtomicState]:=NaturalWidth[state]===0;
SetAttributes[GroundStateQ,Listable];

(*GroundStates[system] returns all ground states.*)

GroundStates[system:{_AtomicState..}]:=Select[system,GroundStateQ];

(*ExcitedStates[system] returns all not ground states.*)

ExcitedStates[system:{_AtomicState..}]:=Select[system,!GroundStateQ[#]&];

SublevelMultiplicity[s_AtomicState]:=
	If[DefinedQ[s,F],
		(2F[s]+1),
		If[DefinedQ[s,J],(2J[s]+1),1]If[DefinedQ[s,NuclearSpin],(2NuclearSpin[s]+1),1]
	]

SublevelMultiplicity[syst:{_AtomicState..}]:=
	Total[SublevelMultiplicity/@(DeleteParameter[SelectState[syst,#],F]&/@Union[StateLabel[syst]])]

(*SelectStates[states, rule1, rule2, ...] returns all states that satisfy the supplied rules*)

(*propMatch[a_->b_,reg_]:=MatchQ[a[reg],b]*)
(*propMatch[expr_,reg_]:=expr/.Rule[a_,b_]:>MatchQ[a[reg],b]
critFunc[crit__][reg_]:=And@@(propMatch[#,reg]&/@{crit})*)
(*SelectStates[system:{_AtomicState..},crit__]:=Select[system,critFunc[crit]]
SelectStates[system:{_AtomicState..},crit__,n_?IntegerQ]:=Take[Select[system,critFunc[crit]],n]*)
(*DeleteStates[system:{_AtomicState..},crit__]:=DeleteCases[system,s_?(critFunc[crit])]*)
(*StatePosition[system:{_AtomicState..},crit__]:=Position[system,s_?(critFunc[crit])]*)
(*StatePosition[system:{_AtomicState..}, crit__] := 
 Position[Transpose@Outer[#1[#2]&, {crit}[[All, 1]], system], {crit}[[All, 2]]]*)

SelectStates[system:{_AtomicState..}, crit__] := Extract[system, StatePosition[system, crit]]

SelectStates[system:{_AtomicState..},crit__,n_?IntegerQ] := Take[SelectStates[system, crit],n]

(*SelectState[states, rule1, rule2, ...] returns the first state that satisfies the supplied rules*)

SelectState[system:{_AtomicState..},crit__] := First[SelectStates[system, crit]]

DeleteStates[system:{_AtomicState..},crit__] := Delete[system,StatePosition[system, crit]]

StateLabelQ[sys:{_AtomicState..}, expr_] := MemberQ[StateLabel[sys], expr]

StatePosition[sys:{_AtomicState..}, crit_]/;StateLabelQ[sys,crit] := StatePosition[sys, StateLabel==crit];

StatePosition[sys:{_AtomicState..}, crit_] := 
	Position[sys, 
  		state_ /; (crit /. Rule -> Equal /. param : (StateParameterPattern[] | StateLabel) -> param[state])
	];

StatePosition[sys:{_AtomicState..}, crits__] := StatePosition[sys, And[crits]]

StateMatchQ[state_AtomicState, crits__] := StatePosition[{state},crits] == {{1}}

(*SetSystemType sets a global value for the AtomicSystemType. This value is used in the functions SubstateLabel, ZeemanSubstates, WignerEckart, OpticalField, E1Hamiltonian, and OpticalRepopulation if not overidden by options set for those functions.*)

AtomicSystemType[state_AtomicState] := Which[
	DefinedQ[state,NuclearSpin],HyperfineZeeman,
	DefinedQ[state,J],Zeeman,
	True,Toy
]

AtomicSystemType[system:{_AtomicState..}]:=AtomicSystemType[First@system]

ASTOpt[opts_,func_,state_]:=AtomicSystemType/.opts/.Options[func]/.Automatic:>AtomicSystemType[state]

ExpandASTOpt[opts_,func_,state_]:=Append[DeleteCases[opts,AtomicSystemType->_],AtomicSystemType->ASTOpt[opts,func,state]]

(*LevelLabel[state] returns a list of the state label and F (for a Hyperfine system).*)

Options[LevelLabel]={AtomicSystemType->Automatic};

LevelLabel[state_AtomicState,opts___]:=Switch[
	ASTOpt[{opts},SublevelLabel,state],
	Toy|Zeeman,StateLabel[state],
	HyperfineZeeman,{StateLabel[state],F[state]}
]

(*SublevelLabel[state] returns a list of the state label, F (for a Hyperfine system) and M (for a Zeeman system).*)

Options[SublevelLabel]={AtomicSystemType->Automatic};

SublevelLabel[state_AtomicState,opts___]:=Switch[
	ASTOpt[{opts},SublevelLabel,state],
	Toy,StateLabel[state],
	Zeeman,{StateLabel[state],M[state]},
	HyperfineZeeman,{StateLabel[state],F[state],M[state]}
]

(*HyperfineSublevels[states] returns the hyperfine substates of the states*)

Options[HyperfineSublevels]={AtomicSystemType->Automatic};

HyperfineSublevels[state_AtomicState,opts___]:=
	If[DefinedQ[state,F],
	state,
	Switch[
	ASTOpt[{opts},HyperfineSublevels,state],
	HyperfineZeeman,Reverse@Table[AddParameter[state,F->f],{f,Abs[NuclearSpin[state]-J[state]],NuclearSpin[state]+J[state]}],
	_,state
	]
]

HyperfineSublevels[system:{_AtomicState..},opts___]:=Module[{opts1=ExpandASTOpt[{opts},HyperfineSublevels,system]},Flatten[HyperfineSublevels[#,Sequence@@opts1]&/@system]]

(*ZeemanSublevels[states]  returns the Zeeman sublevels of the states, if M is not defined and the system type is not Toy. For a Zeeman system, J is used, for a HyperfineZeeman System, F is used.*)

Options[ZeemanSublevels]={AtomicSystemType->Automatic};

ZeemanSublevels[state_AtomicState,opts___]:=
	If[DefinedQ[state,M],
	state,
	Switch[
	ASTOpt[{opts},ZeemanSublevels,state],
	Toy,state,
	Zeeman|HyperfineZeeman,Reverse@Table[AddParameter[state,M->m],{m,-F[state],F[state]}]
]
]

ZeemanSublevels[system:{_AtomicState..},opts___]:=
	Module[{opts1=ExpandASTOpt[{opts},ZeemanSublevels,system]},
		Flatten[ZeemanSublevels[#,Sequence@@opts1]&/@system]
	]

(*Sublevels[states] returns all hyperfine and Zeeman sublevels of the list of states.*)

Sublevels[state_AtomicState]:=ZeemanSublevels[HyperfineSublevels[state]]

Sublevels[system:{_AtomicState..}]:=Flatten[Sublevels/@system]

RemoveZeemanSublevels[syst:{_AtomicState..}]:=First/@Split@DeleteParameter[syst,M]

(*ZeemanSystem[label, J] constructs a spin J atomic system.*)

ZeemanSystem[label_,j_]:=ZeemanSublevels[AtomicState[label,J->j],AtomicSystemType->Zeeman]

ZeemanSystem[j_]:=ZeemanSystem[Unique[],j]

End[] (* End Private Context *)

EndPackage[]