(* Mathematica Package *)

BeginPackage["AtomicDensityMatrix`ExperimentalSystem`", 
	{ "AtomicDensityMatrix`UsageMessages`",
	"AtomicDensityMatrix`Common`",
	"AtomicDensityMatrix`AtomicSystem`"}]

Begin["`Private`"] (* Begin Private Context *) 

(* ::Subsection:: *)
(*Experimental system*)

(*Region[ label ] specifies an experimental region labeled by label*)
(*Region[ label, {\!\(TraditionalForm\`\**)
(*SubscriptBox[*)
(*StyleBox["param",*)
(*FontSlant->"Italic"], "1"]\)->Subscript[value, 1] , ...} ]*)
(*or  Region[ label, \!\(TraditionalForm\`\**)
(*SubscriptBox[*)
(*StyleBox["param",*)
(*FontSlant->"Italic"], "1"]\)->Subscript[value, 1] , ... ] specifies an experimental region with one or more parameters set*)
(**)
(*Region parameters:*)
(*	ElectricField*)
(*	MagneticField*)
(*	OpticalFrequency*)
(*	TransitRate*)
(*	Density*)

Region[label_,{rules___Rule}]:=Region[label,rules]

(*RegionLabel[region] gets the label from a state specification*)

RegionLabel[reg_Region]:=First[reg];
RegionLabel[sys:{_Region..}] := RegionLabel/@sys;

(*RegionParameters[region] gets the parameters from a state specification*)

RegionParameters[reg_Region]^:=Rest[List@@reg]

MapOnParameters[f_,reg_Region]:=Region[RegionLabel[reg],f[RegionParameters[reg]]]

(*RegionParameterPattern[] matches valid experimental parameters*)

RegionParameterPattern[] = OpticalParameters|ElectricField|MagneticField|TransitRate[_]|Density

(*RegionParameterQ[expr] tests if an expression is a valid experimental parameter*)

RegionParameterQ[expr_]:=MatchQ[expr,RegionParameterPattern[]]

(*ExperimentalSystemQ[expr] tests if an expression is a list of Region specifications.*)

ExperimentalSystemQ[expr_]:=MatchQ[expr,{_Region..}]

(*DefinedQ[region,param] checks if an experimental parameter is defined in a Region specification.*)

DefinedQ[region_Region,X:RegionParameterPattern[]]:=(X/.RegionParameters[region])=!=X

DefinedQ[region_Region,X:{RegionParameterPattern[]..}]:=And@@(DefinedQ[region,#]&/@X)

DefinedQ[system:{_Region..},X_]:=And@@(DefinedQ[#,X]&/@system)

(*DeleteParameter[region, param] deletes a rule from a region specification. More than one parameter can be deleted by giving a list of parameters.*)

DeleteParameter[region_Region,X:RegionParameterPattern[]]:=MapOnParameters[DeleteCases[#,X->_]&,region];
DeleteParameter[region_Region,X:{RegionParameterPattern[]..}]:=Fold[DeleteParameter,region,X];
DeleteParameter[system:{_Region..},X_]:=DeleteParameter[#,X]&/@system

(*AddParameter[region, param -> value] adds a rule to a region specification. If the parameter is already present, its value is changed if necessary.*)

AddParameter[region_Region,r:Rule[X:RegionParameterPattern[],value_]]:=MapOnParameters[Append[#,r]&,DeleteParameter[region,X]]

(*param[region] returns the parameter for a state if defined, otherwise returns a placeholder marked by the state label.*)


With[{f=#},
	f[reg_Region] := f[reg] = f /. RegionParameters[reg] /. f->f[RegionLabel[reg]];
	f[sys:{_Region..}] := f/@sys;
]&/@{ElectricField,MagneticField,OpticalParameters,Density};

(*(#[region_Region]:=#/.RegionParameters[region]/.#->#[Label[region]])&/@{ElectricField,MagneticField,OpticalParameters,Density};*)

TransitRate[reg1_Region][reg2_Region]:=TransitRate[RegionLabel[reg1]][reg2]

TransitRate[label_][region_Region]:=
	Module[{value},
		If[
			RegionLabel[region]===label,
			0,
				If[
					(value=TransitRate[label]/.RegionParameters[region])=!=TransitRate[label],
					value,
					TransitRate[RegionLabel[region],label]
				]
		]
	]

(*SelectRegions[regions, rule1, rule2, ...] returns all states that satisfy the supplied rules*)

SelectRegions[es:{_Region..},crit__Rule]:=Select[es,critFunc[crit]]

SelectRegions[es:{_Region..},crit__Rule,n_?IntegerQ]:=Take[Select[es,critFunc[crit]],n]

(*SelectRegion[regions, rule1, rule2, ...] returns the first state that satisfies the supplied rules*)

SelectRegion[es:{_Region..},crit__Rule]:=First[SelectStates[es,crit]]

DeleteRegions[es:{_Region..},crit__Rule]:=DeleteCases[es,_?(critFunc[crit])]

RegionPosition[es:{_Region..},crit__Rule]:=Position[es,_?(critFunc[crit])]

End[] (* End Private Context *)

EndPackage[]